//------------------------------------------------------------------------
// Copyright (C) 2011 Niccolo Rigacci
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// Author:      Niccolo Rigacci <niccolo@rigacci.org>
//
// Version:     2011-03-24
//
// Fake Wake-on-LAN for the LaCie d2 Network 2.
//
// WoL function is missing from the d2net_v2 network adapter, so we
// need to simulate it in software.
//
// Run this program at a very early stage of boot process: it will
// power-down the USB bus, the blue LED and the disk. Then it waits for
// a WoL Magic Packet or a Power Button press/release before to return.
//------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <linux/if_ether.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>

#define BLUE_LED_TRIGGER "/sys/class/leds/d2net_v2:blue:sata/trigger"
#define INPUT_EVENT      "/dev/input/event0"
#define IFNAME           "eth0"
#define EV_SYN           0x00
#define EV_KEY           0x01
#define KEY_RESERVED     0x00
#define KEY_POWER        0x74

// The Linux event structure
struct input_event {
    struct timeval time;
    __u16 type;
    __u16 code;
    __s32 value;
};

//------------------------------------------------------------------------
// Write a string to a file.
//------------------------------------------------------------------------
void write_to(char* file, char* str) {
    FILE *fp;
    fp = fopen(file, "w");
    if (! fp) {
        perror("Error: fopen()");
    } else {
        if (fprintf(fp, str) < 0) {
            printf("Error writing %s to %s\n", str, file);
        }
        fclose(fp);
    }
}

//------------------------------------------------------------------------
// Execute an external command and wait for it to terminate.
//------------------------------------------------------------------------
void fork_exec(char** cmd) {
    pid_t pid;
    int child_exit_status, i;
    pid = fork();
    switch (pid) {
        case -1:
            perror("Error: fork()");
            exit(1);
            break;
        case 0:
            // The child execute a command.
            //printf("Executing command: %s\n", cmd[0]);
            execvp(cmd[0], cmd);
            // Should not return. In case, an error occurred.
            perror("Error: execlp()");
            exit(1);
            break;
        default:
            // The parent: wait the child to terminate.
            waitpid(pid, &child_exit_status, 0);
            //printf("The child returned %d\n", child_exit_status);
            break;
     }
}

//------------------------------------------------------------------------
// Read an input event from fd.
// Return 1 if it is an EV_KEY, KEY_POWER release.
//------------------------------------------------------------------------
int get_power_button_event(int fd) {

    struct input_event event;

    switch (read(fd, &event, sizeof(event))) {
        case -1:
            perror("Error event read()");
        case 0:
            printf("Input event EOF\n");
        default:
            printf("Received input event (type, code, value) = (0x%02X, 0x%02X, 0x%04lX)\n",
                event.type, event.code, event.value);
            if (event.type == EV_KEY && event.code == KEY_POWER && event.value == 0) return 1;
    }
    return 0;
}

//------------------------------------------------------------------------
// Read a raw Ethernet packet from sockfd.
// Return 1 if it is a WoL Magic Packet for our MAC address.
//------------------------------------------------------------------------
int get_magic_packet(int sockfd, char* mac_address) {

    int i, j, k;
    int length;
    int magic_packet_min_length = 6 * 2 + 6 + 6 * 16;
    char* buffer = (void*)malloc(ETH_FRAME_LEN);  // Buffer for ethernet frame
    int found, mac_match;

    length = recvfrom(sockfd, buffer, ETH_FRAME_LEN, 0, NULL, NULL);
    if (length == -1) {
        printf("Error recvfrom()\n");
    } else if (length >= magic_packet_min_length) {
        //for (k = 0; k < length; k++) { printf("%02X", buffer[k]); } printf("\n");
        // Search for the Magic Packet sync sequence (0xff * 6).
        // Skip dst and src MAC addresses (12 bytes) at the begin of Ethernet frame.
        for (i = 12; i <= (length - (6 * 17)); i++) {
            if (buffer[i] == 0xff) {
                found = 1;
                for (j = 1; j < 6; j++) {
                    if (buffer[i + j] != 0xff) {
                        found = 0;
                        break;
                    }
                }
                if (found) {
                    //printf("Received %d bytes\n", length);
                    //printf("Found magic packet sync at offset %d\n", i);
                    mac_match = 1;
                    for (j = 1; j <= 16; j++) {
                        if (memcmp(buffer + (i + j * 6), mac_address, 6) != 0) {
                            mac_match = 0;
                            break;
                        }
                    }
                    if (mac_match) return 1;
                }
            }
        }
    }
    return 0;
}

//------------------------------------------------------------------------
// Main loop.
//------------------------------------------------------------------------
void main(void) {

    char* hdparm_sleep[4] = { "/sbin/hdparm", "-Y",  "/dev/sda", NULL };
    char* ifup[4]         = { "/sbin/ifconfig", IFNAME, "up", NULL};

    int i, k;
    int wol = 0, wob = 0;
    char my_mac[6];             // Store our MAC address.
    struct ifreq ifr;           // Store network interface data.
    struct timespec t;          // Time structure for timeouts.
    fd_set readfds;             // File descripts set to read from.
    int fd, sockfd;             // File and socket descriptor.

    // Get our MAC address for eth0.
    sprintf(ifr.ifr_name, IFNAME);
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd < 0) { perror("Error getting eth0 MAC address: socket()"); exit(1); }
    i = ioctl(sockfd, SIOCGIFHWADDR, (char *)&ifr);
    if (i < 0) { perror("Error getting eth0 MAC address: ioctl()"); exit(1); }
    memcpy(my_mac, ifr.ifr_ifru.ifru_hwaddr.sa_data, 6);
    printf("%s MAC address is: ", IFNAME);
    for (k = 0; k < 6; k++) { printf("%02X", my_mac[k]); } printf("\n");
    close(sockfd);

    // Suspend USB power using the dedicated GPIO port #28 (d2net_v2).
    write_to("/sys/class/gpio/export", "28");
    write_to("/sys/class/gpio/gpio28/direction", "out");
    write_to("/sys/class/gpio/gpio28/value", "0");
    write_to("/sys/class/gpio/unexport", "28");

    // Turn off the blue led.
    write_to(BLUE_LED_TRIGGER, "default-on");
    t.tv_sec = 0;
    t.tv_nsec = 100000000; // 0.1 sec
    nanosleep(&t, NULL);
    write_to(BLUE_LED_TRIGGER, "none");

    // Open a raw socket to listen for Magic Packet.
    fork_exec(ifup);
    sockfd = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
    if (sockfd == -1) {
        perror("Error socket()\n");
        exit(1);
    }

    // Open input to get power button press.
    fd = open(INPUT_EVENT, O_RDONLY);
    if (fd == -1) {
        perror("Error open(\"/dev/input/eventX\")");
        exit(1);
    }

    // Force the drive to immediately enter the lowest power
    // consumption sleep mode, causing it to shut down completely.
    sync();
    t.tv_sec = 2; // 1 sec
    t.tv_nsec = 0;
    nanosleep(&t, NULL);
    fork_exec(hdparm_sleep);

    while (!wol && !wob) {

        // Wait untill data is received from the socket or from the input.
        FD_ZERO(&readfds);
        FD_SET(sockfd, &readfds);
        FD_SET(fd, &readfds);
        i = (fd > sockfd ? fd : sockfd) + 1;
        select(i, &readfds, NULL, NULL, NULL);

        // Check for Power Button event or WoL MagicPacket.
        if (FD_ISSET(fd, &readfds))     wob = get_power_button_event(fd);
        if (FD_ISSET(sockfd, &readfds)) wol = get_magic_packet(sockfd, my_mac);
    }

    close(sockfd);
    close(fd);

    if (wol) printf("Received Wake-on-LAN Magic Packet.\n");
    if (wob) printf("Received EV_KEY, KEY_POWER input event.\n");

    // Wake-up disk
    // The Linux IDE  driver will  automatically  handle
    // issuing a reset if/when needed.

    // Power-ON USB using the dedicated GPIO port #28 (d2net_v2).
    write_to("/sys/class/gpio/export", "28");
    write_to("/sys/class/gpio/gpio28/direction", "out");
    write_to("/sys/class/gpio/gpio28/value", "1");
    write_to("/sys/class/gpio/unexport", "28");

    // Turn on the blue led.
    write_to(BLUE_LED_TRIGGER, "default-on");

}
